#include "util.h"
#include "gsm.h"
#include "gsm_private.h"


const GBFS_FILE *gsm_fs;


static struct gsm_state gsm_decoder;
static const char *gsm_src_start = NULL;
static const char *gsm_src_pos = NULL;
static const char *gsm_src_end = NULL;
static const char *gsm_loop_start = NULL;
static unsigned int gsm_decode_pos = 160, gsm_cur_buffer = 0;

signed short out_samples[160];
signed char double_buffers[2][304] __attribute__ ((aligned (4)));


u32 the_sound[4] =
{0x67452301,0xEFCDAB89,0x98BADCFE,0x10325476};


void update_waveform(u32 period)
{
  unsigned char tri_wavedata[16];
  const u32 *srcL = (const u32 *)tri_wavedata;
  u32 i;

  /* convert wave data */
  for(i = 0; i < 16; i++)
    tri_wavedata[i] = (the_sound[2*i] << 4) | the_sound[2*i+1];

  /* copy wave data */
  srcL = (const u32 *)tri_wavedata;
  TRICTRL = TRICTRL_1X64 | TRICTRL_BANK(1) | TRICTRL_ENABLE;
  TRIWAVERAM[0] = *srcL++;
  TRIWAVERAM[1] = *srcL++;
  TRIWAVERAM[2] = *srcL++;
  TRIWAVERAM[3] = *srcL++;
  TRICTRL = TRICTRL_1X64 | TRICTRL_BANK(0) | TRICTRL_ENABLE;

  TRIFREQ = period | FREQ_HOLD;
  TRILENVOL = TRILENVOL_100;
}


static void dsound_switch_buffers(const void *src, unsigned int ch)
{
  DMA[ch+1].control = 0;

  /* no-op to let DMA registers catch up */
  asm volatile ("eor r0, r0; eor r0, r0" ::: "r0");

  DMA[ch+1].src = src;
  DMA[ch+1].dst = (void *)(ch ? 0x040000a4 : 0x040000a0);
  DMA[ch+1].count = 1;
  DMA[ch+1].control = DMA_DSTUNCH | DMA_SRCINC | DMA_REPEAT | DMA_U32 |
                      DMA_SPECIAL | DMA_ENABLE;
}

void dsound_switch_in(void)
{
  dsound_switch_buffers(double_buffers[gsm_cur_buffer], 0);
}


void init_sound(void)
{
  TIMER[0].control = 0;
  //turn on sound circuit
  SETSNDRES(1);
  SNDSTAT = SNDSTAT_ENABLE;
  DSOUNDCTRL = 0x0b06;
  DMGSNDCTRL = 0xff77;
  TIMER[0].count = 0x10000 - 924;
  TIMER[0].control = TIMER_16MHZ | TIMER_ENABLE;
  update_waveform(0);
}



__attribute__((long_call))
void gsm_init (gsm r);

void gsm_pass(void)
{
  u.n_passes++;    
  if(gsm_loop_start)  /* if we've already passed the beginning of the loop */
  {
    if(u.n_passes == u.cur_song.music_loop_len)
    {
      gsm_src_pos = gsm_loop_start;
      u.n_passes = 0;
    }
  }
  else
  {
    if(u.n_passes == u.cur_song.music_loop_start)
    {
      gsm_loop_start = gsm_src_pos;
      u.n_passes = 0;
    }
  }
}

void gsm_playback(const char *new_song)
{
  unsigned int j;
  signed char *dst_pos;
  unsigned int need_new_song = new_song ? 1 : 0;

  if(!gsm_fs)
    return;

  gsm_cur_buffer = !gsm_cur_buffer;
  dst_pos = double_buffers[gsm_cur_buffer];

  if(gsm_src_pos >= gsm_src_end)
    need_new_song = 1;
  if(need_new_song)
  {
    const char *src;
    unsigned int src_len;

    gsm_loop_start = NULL;
    u.n_passes = 0;
    u.gsm_volume = 256;

    gsm_init(&gsm_decoder);
    src = gbfs_get_obj(gsm_fs, new_song, &src_len);
    gsm_src_start = src;
    gsm_src_pos = src;
    gsm_src_end = src + src_len;
    gsm_decode_pos = 160;
  }

  for(j = 304 / 4; j > 0; j--)
  {
    unsigned int vol = u.gsm_volume;
    if(gsm_decode_pos >= 160)
      {
        if(gsm_src_pos < gsm_src_end)
 	  gsm_decode(&gsm_decoder, gsm_src_pos, out_samples);
	gsm_src_pos += sizeof(gsm_frame);
	gsm_decode_pos = 0;
      }
    dst_pos[0] = out_samples[gsm_decode_pos + 0] * vol >> 16;
    dst_pos[1] = out_samples[gsm_decode_pos + 1] * vol >> 16;
    dst_pos[2] = out_samples[gsm_decode_pos + 2] * vol >> 16;
    dst_pos[3] = out_samples[gsm_decode_pos + 3] * vol >> 16;

    gsm_decode_pos += 4;
    dst_pos += 4;
  }
}


