/* mkskin.c
   make a luminesweeper game left skin file

Copyright 2005 Damian Yerrick

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to 
  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA  02111-1307, USA.
GNU licenses can be viewed online at http://www.gnu.org/copyleft/

Visit http://www.pineight.com/ for more information.

*/

#if 1
#define USE_CONSOLE
#endif

#include <allegro.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>


BITMAP *bg, *inner_bg, *pf, *pf_tiles;


#define GB_PERIOD(x) (2048 - (int)(131072 / (x)))

char music_filename[32];
unsigned long int music_bpm;
unsigned int music_loop_start;
unsigned int music_loop_len;
unsigned int move_period = GB_PERIOD(1244.51 * 1.5);
unsigned int turn_period = GB_PERIOD(1244.51);
unsigned int blawk_period = GB_PERIOD(77.78);

/* pf_tile_coords
______________________________________________________________________
Each row is the x and y coordinates (in tiles, not pixels), starting
from top left, of tile id n in the file.

 0-15: H blocks
16-31: R blocks
32   : bomb overlay
33   : swept block
34-35: sweeper
36-37: unused
38-47: digits

*/
#define N_PF_TILES 48
#define PF_TILE_WID 12
#define PF_TILE_HT 12
const unsigned char pf_tile_coords[N_PF_TILES][2] =
{
  { 6, 4}, { 4, 3}, { 1, 3}, { 3, 9},
  { 3, 1}, { 3, 4}, { 2, 5}, { 3, 3},
  { 2, 1}, { 2, 7}, { 2, 4}, { 2, 3},
  { 3, 7}, { 3, 2}, { 2, 2}, { 3, 8},

  { 1, 4}, { 9, 3}, { 6, 3}, { 8, 9},
  { 8, 1}, { 8, 4}, { 7, 5}, { 8, 3},
  { 7, 1}, { 7, 7}, { 7, 4}, { 7, 3},
  { 8, 7}, { 8, 2}, { 7, 2}, { 8, 8},

  { 0, 9}, {11, 8}, {12, 0}, {12, 1},
  { 0, 0}, { 0, 0}, { 2, 0}, { 3, 0},
  { 4, 0}, { 5, 0}, { 6, 0}, { 7, 0},
  { 8, 0}, { 9, 0}, {10, 0}, {11, 0}

};

typedef unsigned long int rgbcolor_t;

rgbcolor_t bg_pal[16];
int bg_pal_len;
rgbcolor_t inner_bg_pal[16];
int inner_bg_pal_len;
rgbcolor_t h_pal[16];
int h_pal_len;
rgbcolor_t r_pal[16];
int r_pal_len;
rgbcolor_t bomb_pal[16];
int bomb_pal_len;
rgbcolor_t swept_pal[16];
int swept_pal_len;
rgbcolor_t sweeper_pal[16];
int sweeper_pal_len;
rgbcolor_t digits_pal[16];
int digits_pal_len;

unsigned char skinfile[65536];



/* parse_ini_line *******************
*/
int parse_ini_line(const char *in, char *key, char *var, char *val)
{
  int c;
  char *kstart = key;

  /* Skip whitespace before key */
  while(*in && isspace(*in))
    in++;

  /* Parse key */
  if(*in == '[')  /* if we have a new key, load it */
  {
    in++;

    /* Skip whitespace before actual key */
    while(*in && isspace(*in))
      in++;

    for(c = *in++;
        c != 0 && c != ']' && c != '\n' && c != '\r';
        c = *in++)
    {
      if(!isspace(c))
        *key++ = c;
    }
    *key = 0;
    /* Strip whitespace after key */
    do {
      *key-- = 0;
    } while(key >= kstart && isspace(*key));
  }

  /* Skip whitespace before variable */
  while(*in && isspace(*in))
    in++;
  if(*in == 0)  /* if there is no variable, don't do anything */
    return 0;

  for(c = *in++;
      c != 0 && c != '=' && c != '\n' && c != '\r';
      c = *in++)
  {
    if(!isspace(c))
    {
      *var++ = c;
    }
  }
  *var = 0;

  /* Skip whitespace before value */
  while(*in && isspace(*in))
    in++;

  /* Get value */
  kstart = val;
  for(c = *in++;
      c != 0 && c != '\n' && c != '\r';
      c = *in++)
  {
    *val++ = c;
  }
  /* Strip whitespace after value */
  do {
    *val-- = 0;
  } while(val >= kstart && isspace(*val));

  return 0;
}


int trim_to_pf(BITMAP **in)
{
  fputs("trimming\n", stdout);
  if(in && *in)
  {
    BITMAP *new_bmp = create_bitmap(192, 120);

    if(new_bmp)
    {
      blit(*in, new_bmp,
           32, 32,
           0, 0,
           192, 160);
      destroy_bitmap(*in);
      *in = new_bmp;
      return 0;
    }
    else
      return -1;
  }
  else
    return -1;
}


/* parse_ini_file() *****************
   Parses the INI file and loads bitmaps.
   Returns 0 for success or <0 for failure.
*/
int parse_ini_file(FILE *fp)
{
  char key[1024], var[1024], val[1024], input_buf[1024];

  key[0] = 0;
  var[0] = 0;
  val[0] = 0;

  while(1)
  {
    int rval;

    if(!fgets(input_buf, sizeof(input_buf), fp))
      break;
    rval = parse_ini_line(input_buf, key, var, val);

    if(!strcmp("BG", var))
    {
      bg = load_bitmap(val, NULL);
      if(!bg)
      {
        allegro_message("Could not load background image from %s\n",
                         val);
        return -1;
      }
      if(bg->w != 240 || bg->h != 160)
      {
        allegro_message("Background image is %ux%u pixels, not 240x160 as expected.\n",
                         bg->w, bg->h);
        return -1;
      }
    }
    else if(!strcmp("InnerBG", var))
    {
      inner_bg = load_bitmap(val, NULL);
      if(!bg)
      {
        allegro_message("Could not load inner background image from %s\n",
                         val);
        return -1;
      }
      if(inner_bg->w == 240 && inner_bg->h == 160
         && trim_to_pf(&inner_bg) < 0)
      {
        allegro_message("Could not trim inner background image %s\n",
                         val);
        return -1;
      }
      if(inner_bg->w != 192 || inner_bg->h != 120)
      {
        allegro_message("Inner background image is %ux%u pixels, not 240x160 or 192x120 as expected.\n",
                         inner_bg->w, inner_bg->h);
        return -1;
      }
    }
    else if(!strcmp("PF", var))
    {
      pf = load_bitmap(val, NULL);
      if(!pf)
      {
        allegro_message("Could not load inner background image from %s\n",
                         val);
        return -1;
      }
      if(pf->w == 240 && pf->h == 160 && trim_to_pf(&pf) < 0)
      {
        allegro_message("Could not trim playfield image %s\n",
                         val);
        return -1;
      }
      if(pf->w != 192 || pf->h != 120)
      {
        allegro_message("Playfield image is %ux%u pixels, not 240x160 or 192x120 as expected.\n",
                         pf->w, pf->h);
        return -1;
      }
    }
    else if(!strcmp("Music", var))
    {
      strncpy(music_filename, val, 25);
    }
    else if(!strcmp("MusicBPM", var))
    {
      double bpm = atof(val);

      music_bpm = 9363.2825980797121404049861394136 * bpm;
    }
    else if(!strcmp("MusicLoopStart", var))
    {
      music_loop_start = atoi(val);
    }
    else if(!strcmp("MusicLoopLen", var))
    {
      music_loop_len = atoi(val);
    }
    else if(!strcmp("MoveFreq", var))
    {
      double bpm = atof(val);

      move_period = GB_PERIOD(bpm);
    }
    else if(!strcmp("RotateFreq", var))
    {
      double bpm = atof(val);

      turn_period = GB_PERIOD(bpm);
    }
    else if(!strcmp("BlawkFreq", var))
    {
      double bpm = atof(val);

      blawk_period = GB_PERIOD(bpm);
    }
    else
      allegro_message("Unknown key [%s]%s=%s(got: %d)\n", key, var, val, rval);
  }
  return 0;
}


/* get_16_colors() **********************
   Gets the first 16 colors from a rectangle whose upper left corner
   is (left, top) and whose size is (w, h) pixels.
   Reserves the first n existing colors in the palette.
   Returns the number of colors, or -1 if there are more than 16
   unique colors.
*/
int get_16_colors(BITMAP *src, rgbcolor_t *dst,
                  unsigned int left, unsigned int top,
                  unsigned int w, unsigned int h,
                  unsigned int n)
{
  unsigned int i, x;

  for(; h > 0; top++, h--)
  {
    for(x = 0; x < w; x++)
    {
      unsigned int c = getpixel(src, x + left, top);

      /* Search for this color. */
      for(i = 0; i < n; i++)
      {
        if(c == dst[i])
          goto seen_this_color_before;
      }

      /* If there's space, add this color. */
      if(n >= 16)
        return -1;
      dst[n++] = c;
seen_this_color_before:
      (void)0;
    }
  }
  return n;
}

void draw_16_colors(BITMAP *bmp, const rgbcolor_t *pal,
                    unsigned int left, unsigned int top)
{
  unsigned int x;

  acquire_bitmap(bmp);
  for(x = 0; x < 16; x++)
    rectfill(bmp,
             x * 8 + left, top, x * 8 + left + 7, top + 7,
             pal[x]);
  release_bitmap(bmp);
}

int make_common_color(rgbcolor_t *pal1, rgbcolor_t *pal2)
{
  unsigned int i1, i2;

  for(i1 = 0; i1 < 16; i1++)
    for(i2 = 0; i2 < 16; i2++)
      if(pal1[i1] == pal2[i2])
      {
        unsigned int t;

        t = pal1[i1]; pal1[i1] = pal1[0]; pal1[0] = t;
        t = pal2[i2]; pal2[i2] = pal2[0]; pal2[0] = t;
        return 0;
      }
  return -1;
}


void unclut_bitmap(BITMAP *src, rgbcolor_t *pal,
                   unsigned int left, unsigned int top,
                   unsigned int w, unsigned int h,
                   unsigned int n)
{
  unsigned int x, i;

  for(; h > 0; top++, h--)
  {
    for(x = 0; x < w; x++)
    {
      unsigned int c = getpixel(src, x + left, top);

      /* Search for this color. */
      for(i = 0; i < n; i++)
      {
        if(c == pal[i])
        {
          c = i;
          goto right_c;
        }
      }
      c = 0;
right_c:
      putpixel(src, x + left, top, c);
    }
  }
}


unsigned long int pack_palette(const rgbcolor_t *src, unsigned char *dst)
{
  unsigned int i;

  for(i = 0; i < 16; i++)
  {
    unsigned int c = *src++;
    unsigned int r = getr(c);
    unsigned int g = getg(c);
    unsigned int b = getb(c);

    r = ((int)sqrt(r * 256) + r) >> 4;
    g = ((int)sqrt(g * 256) + g) >> 4;
    b = ((int)sqrt(b * 256) + b) >> 4;

    c = r | (g << 5) | (b << 10);
    *dst++ = c & 0xff;
    *dst++ = (c >> 8) & 0xff;
  }
  return 32;
}


unsigned long int pack_tile(BITMAP *src, unsigned char *dst,
                            unsigned int left, unsigned int y,
                            unsigned int w, unsigned int h)
{
  unsigned int x;
  unsigned char *const dst_old = dst;

  for(; h > 0; h--, y++)
    for(x = left; x < left + w; x += 2)
    {
      unsigned int c1 = getpixel(src, x, y) & 0x000f;
      unsigned int c2 = getpixel(src, x + 1, y) & 0x000f;
      *dst++ = c1 | (c2 << 4);
    }

  return dst - dst_old;
}


unsigned long int pack_u16(unsigned int n, unsigned char *dst)
{
  dst[0] = n;
  dst[1] = n >> 8;
  return 2;
}


unsigned long int pack_u32(unsigned long int n, unsigned char *dst)
{
  dst[0] = n;
  dst[1] = n >> 8;
  dst[2] = n >> 16;
  dst[3] = n >> 24;
  return 4;
}


unsigned long int pack_everything(void)
{
  unsigned long int offset = 0;
  unsigned int x, y;

  offset += pack_palette(bg_pal,       skinfile+offset);
  offset += pack_palette(inner_bg_pal, skinfile+offset);
  offset += pack_palette(h_pal,        skinfile+offset);
  offset += pack_palette(r_pal,        skinfile+offset);
  offset += pack_palette(bomb_pal,     skinfile+offset);
  offset += pack_palette(swept_pal,    skinfile+offset);
  offset += pack_palette(sweeper_pal,  skinfile+offset);
  offset += pack_palette(digits_pal,   skinfile+offset);

  for(y = 0; y < 160; y += 8)
    for(x = 0; x < 240; x += 8)
      offset += pack_tile(bg, skinfile+offset, x, y, 8, 8);
  for(y = 0; y < 36; y += 12)
    for(x = 0; x < 192; x += 12)
      offset += pack_tile(pf_tiles, skinfile+offset, x, y, 12, 12);

  strncpy(skinfile+offset, music_filename, 24);
  offset += 24;

  offset += pack_u32(music_bpm,        skinfile+offset);
  offset += pack_u16(music_loop_start, skinfile+offset);
  offset += pack_u16(music_loop_len,   skinfile+offset);
  offset += pack_u16(move_period,      skinfile+offset);
  offset += pack_u16(turn_period,      skinfile+offset);
  offset += pack_u16(blawk_period,     skinfile+offset);
  offset += pack_u16(0,                skinfile+offset);

  return offset;
}


int main(const int argc, const char *const *const argv)
{
  FILE *fp;
  install_allegro(SYSTEM_NONE, &errno, atexit);
  set_color_depth(32);

  if(argc != 3)
  {
    allegro_message("wrong number of arguments; try mkskin kitten.ini kitten.Lskin\n");
    return EXIT_FAILURE;
  }

  fp = fopen(argv[1], "rt");
  if(!fp)
  {
    allegro_message("Could not open skin description file %s: %s\n",
                    argv[1], strerror(errno));
    return EXIT_FAILURE;
  }

  /* First parse the INI file and load all bitmap images */
  if(parse_ini_file(fp) < 0)
  {
    fclose(fp);
    allegro_message("Could not parse %s\n", argv[1]);
    return EXIT_FAILURE;
  }
  fclose(fp);
  fp=NULL;

  /* Put inner_bg in a separate bitmap if it isn't already */
  if(!inner_bg)
  {
    inner_bg = create_bitmap(192, 120);
    if(!inner_bg)
    {
      allegro_message("Not enough memory for the inner background\n");
      return EXIT_FAILURE;
    }
    blit(bg, inner_bg, 32, 32, 0, 0, 192, 120);
  }

  /* Clear out the part of outer background that's covered by the
     inner background */
  rectfill(bg, 32, 32, 191+32, 119+32, getpixel(bg, 0, 0));

  /* Rearrange bg tiles from template into order used by GBA */
  pf_tiles = create_bitmap(192, 120);
  if(!pf_tiles)
  {
    allegro_message("Not enough memory for playfield tiles\n");
    return EXIT_FAILURE;
  }
  clear_bitmap(pf_tiles);
  {
    int i;

    for(i = 0; i < N_PF_TILES; i++)
    {
      unsigned int src_x = PF_TILE_WID * pf_tile_coords[i][0];
      unsigned int src_y = PF_TILE_HT * pf_tile_coords[i][1];
      unsigned int dst_x = (i & 0x0f) * PF_TILE_WID;
      unsigned int dst_y = (i >> 4) * PF_TILE_HT;

      blit(pf, pf_tiles, src_x, src_y, dst_x, dst_y, PF_TILE_WID, PF_TILE_HT);
    }
  }

  /* Get background palettes */
  bg_pal_len = get_16_colors(bg, bg_pal, 0, 0, 240, 160, 0);
  if(bg_pal_len < 0)
  {
    allegro_message("There are more than 16 colors in outer background.\n");
    return EXIT_FAILURE;
  }

  inner_bg_pal_len = get_16_colors(inner_bg, inner_bg_pal, 0, 0, 192, 120, 0);
  if(inner_bg_pal_len < 0)
  {
    allegro_message("There are more than 16 colors in inner background.\n");
    return EXIT_FAILURE;
  }

  /* Get playfield palettes */
  h_pal[0] = getpixel(pf, 0, 0);
  r_pal[0] = getpixel(pf, 0, 0);
  bomb_pal[0] = getpixel(pf, 0, 0);
  swept_pal[0] = getpixel(pf, 0, 0);
  sweeper_pal[0] = getpixel(pf, 0, 0);
  digits_pal[0] = getpixel(pf, 0, 0);

  h_pal_len = get_16_colors(pf_tiles, h_pal, 0, 0, 192, 12, 1);
  if(h_pal_len < 0)
  {
    allegro_message("There are more than 15 nontransparent colors in H blocks.\n");
    return EXIT_FAILURE;
  }

  r_pal_len = get_16_colors(pf_tiles, r_pal, 0, 12, 192, 12, 1);
  if(r_pal_len < 0)
  {
    allegro_message("There are more than 15 nontransparent colors in R blocks.\n");
    return EXIT_FAILURE;
  }

  bomb_pal_len = get_16_colors(pf_tiles, bomb_pal, 0, 24, 12, 12, 1);
  if(bomb_pal_len < 0)
  {
    allegro_message("There are more than 15 nontransparent colors in the Bomb.\n");
    return EXIT_FAILURE;
  }

  swept_pal_len = get_16_colors(pf_tiles, swept_pal, 12, 24, 12, 12, 1);
  if(swept_pal_len < 0)
  {
    allegro_message("There are more than 15 nontransparent colors in the swept block.\n");
    return EXIT_FAILURE;
  }

  sweeper_pal_len = get_16_colors(pf_tiles, sweeper_pal, 24, 24, 24, 12, 1);
  if(sweeper_pal_len < 0)
  {
    allegro_message("There are more than 15 nontransparent colors in the sweeper.\n");
    return EXIT_FAILURE;
  }

  digits_pal_len = get_16_colors(pf_tiles, digits_pal, 72, 24, 120, 12, 1);
  if(digits_pal_len < 0)
  {
    allegro_message("There are more than 15 nontransparent colors in the digits.\n");
    return EXIT_FAILURE;
  }

/* Divide the background palette into backdrop (1), outer palette (15),
   and inner palette (15).
*/
  if(bg_pal_len == 16 && inner_bg_pal_len == 16)
  {
    if(make_common_color(inner_bg_pal, bg_pal) < 0)
      allegro_message("There are more than 31 colors in the background.\n");
  }
  else if(inner_bg_pal_len < 16 && bg_pal_len == 16)
  {
    inner_bg_pal[inner_bg_pal_len++] = inner_bg_pal[0];
    inner_bg_pal[0] = bg_pal[0];
  }
  else
  {
    bg_pal[bg_pal_len++] = bg_pal[0];
    bg_pal[0] = inner_bg_pal[0];
  }

#if 0
  draw_16_colors(pf_tiles, bg_pal, 0, 48);
  draw_16_colors(pf_tiles, inner_bg_pal, 0, 56);
  draw_16_colors(pf_tiles, h_pal, 0, 64);
  draw_16_colors(pf_tiles, r_pal, 0, 72);
  draw_16_colors(pf_tiles, bomb_pal, 0, 80);
  draw_16_colors(pf_tiles, swept_pal, 0, 88);
  draw_16_colors(pf_tiles, sweeper_pal, 0, 96);
  draw_16_colors(pf_tiles, digits_pal, 0, 104);
  save_bitmap("tiles. (period).bmp", pf_tiles, NULL);
#endif

  /* Convert each to 16 colors using the palettes learned earlier. */
  unclut_bitmap(bg, bg_pal, 0, 0, 240, 160, bg_pal_len);
  unclut_bitmap(inner_bg, inner_bg_pal, 0, 0, 192, 120, inner_bg_pal_len);
  unclut_bitmap(pf_tiles, h_pal, 0, 0, 192, 12, h_pal_len);
  unclut_bitmap(pf_tiles, r_pal, 0, 12, 192, 12, r_pal_len);
  unclut_bitmap(pf_tiles, bomb_pal, 0, 24, 12, 12, bomb_pal_len);
  unclut_bitmap(pf_tiles, swept_pal, 12, 24, 12, 12, swept_pal_len);
  unclut_bitmap(pf_tiles, sweeper_pal, 24, 24, 24, 12, sweeper_pal_len);
  unclut_bitmap(pf_tiles, digits_pal, 72, 24, 120, 12, digits_pal_len);

  /* Compose background. */
  blit(inner_bg, bg, 0, 0, 32, 32, 192, 120);

  {
    unsigned long int len = pack_everything();

    fp = fopen(argv[2], "wb");
    if(!fp)
    {
      allegro_message("Could not open %s for writing: %s\n",
                      argv[2], strerror(errno));
      return EXIT_FAILURE;
    }

    fwrite(skinfile, 1, len, fp);
    fclose(fp);
  }


  return 0;
} END_OF_MAIN();



/*

Quadra - Beginning Of Life replaces lumines::Urbanization

*/
