/* util.h
   utilities for luminesweeper

Copyright 2005 Damian Yerrick

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author contact information is in readme.txt

*/
#ifndef INCLUDE_UTIL_H
#define INCLUDE_UTIL_H

#include <stdlib.h>
#include "pin8gba.h"
#include "gbfs.h"

typedef s32 fixed;

typedef struct SONG_SPEC
{
  char name[24];
  u32 music_speed;
  u16 music_loop_start;
  u16 music_loop_len;
  u16 move_period;
  u16 turn_period;
  u16 blawk_period;
  u16 reserved_period;
} SONG_SPEC;


typedef struct Globals
{
  struct OAM_SPRITE oam[128];
  SONG_SPEC cur_song;
  unsigned char n_sprites;
  unsigned char state, state_time;
  unsigned char paused;
  unsigned int retraces;
  unsigned int sweep_sub;
  unsigned short n_passes;
  unsigned char repeats[10];
  unsigned int next_skin_at;
  const void *cur_skin_data;
  unsigned int gsm_volume;
} UtilGlobals;

typedef unsigned char LuBlock;

#define LU_SCORE_ADD_N 4

typedef struct LScoreAdd {
  unsigned char start_x, start_y, countdown, reserved;
  unsigned int value;
} LuScoreAdd;

typedef struct LGlobals
{
  LuBlock b[10][16];
  LuScoreAdd score_add[LU_SCORE_ADD_N];
  struct BGPOINT hdma[160][2];
  unsigned char piece[4];
  unsigned char destroyer_x[4], destroyer_y[4];
  unsigned char sweep_x;
  unsigned char sweepgrp_left;
  unsigned char sweepgrp_deletes;
  unsigned char deletes_this_pass;  /* can never be bigger than 15*9=105 */
  unsigned char deletes_last_pass, deletes_move_cd;
  unsigned char skin_transition;
  unsigned char x;
  unsigned char x_pixel;
  unsigned short y_speed;
  fixed y;
  unsigned short dirty_columns;
  unsigned short floaty_columns;
  unsigned int score, deletes, n_pieces;
  unsigned char cur_skin;
} LGlobals;

typedef struct RGlobals
{
  unsigned char pf[16][32];
  unsigned char reveal_yp2[32];
  unsigned char reveal_yp1[32];
  unsigned char pf_left, pf_top, pf_right, pf_bottom;
  unsigned short n_mines, n_flags, n_spaces, dirty_rows;
} RGlobals;

typedef union
{
  LGlobals left;
  RGlobals right;
} GameSpecificGlobals;

/* menu.c */
extern GameSpecificGlobals game_globals;
extern const GBFS_FILE *data_gbfs;
extern const GBFS_FILE *Lskins_gbfs;

/* util.c */
#define REPEAT_DELAY 10

extern UtilGlobals u;

void clear_keybuf(void);
void make_repeats(u32 keys);
void dma_memcpy(void *dst, const void *src, size_t len);
void nttextout(u32 nt, u32 x, u32 y, u32 c, const char *str);
void itoa_lpad(char *buf, size_t len, unsigned int n, int lpad_chr);
void expand_gb_to_gba4bit(void *in_dst, const u8 *src, size_t len, const unsigned char gbpal[4]);
void load_todbg(unsigned int slot, const unsigned char *name);
void set_palette_dimmed(void *in_dst,
                        const void *in_src,
                        unsigned int dim_to,
                        unsigned int dim_level);

/* asm.s */
__attribute__((long_call))
int dv(int num, int den);
__attribute__((long_call))
void wait4vbl(void);
__attribute__((long_call))
void gblz_unpack(const void *src, void *dst);

/* isr.c */
__attribute__((long_call))
void isr(void);

/* mine.c */
void Rplay(void);

/* lu.c */
void Lplay(void);

/* gsmplay.c */
extern const GBFS_FILE *gsm_fs;
void dsound_switch_in(void);
void init_sound(void);
void gsm_playback(const char *new_song);
void gsm_pass(void);

#endif
